/*
 * Copyright (c) 1996, 1997 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <Dem/usgsdem.h>

#ifndef MAKEDEPEND
#include <vtkDecimatePro.h>
#include <vtkFloatPoints.h>
#include <vtkFloatScalars.h>
#include <vtkPolyDataWriter.h>
#include <vtkStructuredGrid.h>
#include <vtkStructuredGridGeometryFilter.h>
#include <vtkTriangleFilter.h>
#endif

extern "C" {
#include <projects.h>
#include <sys/stat.h>
#include <unistd.h>
}

#include <stdio.h>
#include <string.h>

short round(float x) {
    return x > 0 ? short(x+0.5) : -short(-x+0.5);
}

int main(int argc, char* argv[]) {
    if (argc != 3) {
	printf("usage: usgsdem2smf dem targetdir\n");
	return(-1);
    }

    UsgsDem* dem = new UsgsDem();
    dem->read(argv[1]);

    if (dem->reference_system() == UsgsDem::GEOGRAPHIC &&
	dem->rows() == 1201 && dem->cols() == 1201) {
	// 1 degree USGS DEM
	for (int k = 0; k < 10; k++) {
	    for (int l = 0; l < 10; l++) {
		vtkStructuredGrid* grid = vtkStructuredGrid::New();
		grid->SetDimensions(121, 121, 1);
		vtkFloatPoints* points = vtkFloatPoints::New();
		points->Allocate(121 * 121);
		double dx, dy;
		float p[3];
		for (int j = 0; j < 121; j++)
		    for (int i = 0; i < 121; i++) {
			int di = k * 120 + i;
			int dj = l * 120 + j;
			dem->projected_position(dj, di, dx, dy);
			p[0] = dx;
			p[1] = dy;
			p[2] = dem->value(dj, di);
			points->InsertNextPoint(p);
		    }
		grid->SetPoints(points);
		points->Delete();
		
		vtkStructuredGridGeometryFilter* gfilter = 
		    vtkStructuredGridGeometryFilter::New();
		gfilter->SetExtent(0, 120, 0, 120, 0, 0);
		gfilter->SetInput(grid);
		
		vtkTriangleFilter* tfilter = vtkTriangleFilter::New();
		tfilter->PassVertsOff();
		tfilter->PassLinesOff();
		tfilter->SetInput(gfilter->GetOutput());
		tfilter->Update();

		int flen = strlen(argv[2]) + 80;
		char* fname = new char[flen];
		memset(fname, 0, flen);
		strcpy(fname, argv[2]);
		if (fname[strlen(fname)-1] != '/')
		    strcat(fname, "/");
		const char* demname = dem->Name();
		char c;
		while (c = *demname++)
		    if (!isspace(c))
			fname[strlen(fname)] = c;
		fname[strlen(fname)] = '.';
		fname[strlen(fname)] = '0' + k;
		fname[strlen(fname)] = '.';
		fname[strlen(fname)] = '0' + l;

		strcat(fname, ".smf");
		printf("filename is %s\n", fname);

		// generate smf file
		FILE* smf = fopen(fname, "w");
		for (int ip = 0;
		     ip < tfilter->GetOutput()->GetNumberOfPoints();
		     ip++) {
		  float p[3];
		  tfilter->GetOutput()->GetPoint(ip, p);
		  fprintf(smf, "v %f %f %f\n", p[0], p[1], p[2]);
		}
		int npts, *pts;
		vtkCellArray* cells = tfilter->GetOutput()->GetPolys();
		for (cells->InitTraversal(); cells->GetNextCell(npts, pts); )
		  fprintf(smf, "f %d %d %d\n", pts[0]+1, pts[1]+1, pts[2]+1);
		fclose(smf);

		tfilter->Delete();
		gfilter->Delete();
		grid->Delete();
	    }
	}
    }
    else if (dem->reference_system() == UsgsDem::UTM) {
      // 7.5 minute USGS DEM
      int xdiv = 3;
      int ydiv = 4;
      int xnom = dem->cols() / xdiv;
      int ynom = dem->rows() / ydiv;
      for (int k = 0; k < xdiv; k++) {
	for (int l = 0; l < ydiv; l++) {
	  vtkCellArray* cells = vtkCellArray::New();
	  cells->Allocate(xnom*ynom*2);
	  vtkFloatPoints* fltpoints = NULL;
	  fltpoints = vtkFloatPoints::New();
	  float fltp[3];
	  long xmin, ymin, xmax, ymax;
	  dem->utm_bbox(xmin, ymin, xmax, ymax);
	  int xylen = (xnom+1) * (ynom+1);
	  int* ptidx = new int[xylen];
	  for (int n = 0; n < xylen; n++)
	    ptidx[n] = -1;
	  int idx = 0;
	  for (int i = 0; i < xnom+1; i++) {
	    for (int j = 0; j < ynom+1; j++) {
	      int di = k * xnom + i;
	      int dj = l * ynom + j;
	      if (dem->elevation_units() == UsgsDem::METERS) {
		fltp[0] = xmin + di * 30;
		fltp[1] = ymin + dj * 30;
		fltp[2] = dem->value(dj, di);
		fltpoints->InsertNextPoint(fltp);
	      }
	      else if (dem->elevation_units() == UsgsDem::FEET) {
		fltp[0] = xmin + di * 30;
		fltp[1] = ymin + dj * 30;
		fltp[2] = dem->value(dj, di) * 1.0/3.28;
		fltpoints->InsertNextPoint(fltp);
	      }
	      ptidx[j + (ynom+1) * i] = idx++;
	    }
	  }
	  int intp[3];
	  for (int i = 0; i < xnom; i++) {
	    for (int j = 0; j < ynom; j++) {
	      int di = k * xnom + i;
	      int dj = l * ynom + j;
	      if (di < dem->cols()-1 && dj < dem->rows()-1) {
		if (ptidx[(j+1) + (ynom+1) * i] > 0 &&
		    ptidx[(j+1) + (ynom+1) * (i+1)] > 0 &&
		    ptidx[j + (ynom+1) * i] > 0) {
		  intp[0] = ptidx[j + (ynom+1) * i];
		  intp[2] = ptidx[(j+1) + (ynom+1) * i]; //counter-clock
		  intp[1] = ptidx[(j+1) + (ynom+1) * (i+1)];
		  cells->InsertNextCell(3, intp);
		}
		if (ptidx[j + (ynom+1) * (i+1)] > 0 &&
		    ptidx[(j+1) + (ynom+1) * (i+1)] > 0 &&
		    ptidx[j + (ynom+1) * i] > 0) {
		  intp[0] = ptidx[j + (ynom+1) * i];
		  intp[1] = ptidx[j + (ynom+1) * (i+1)];
		  intp[2] = ptidx[(j+1) + (ynom+1) * (i+1)];
		  cells->InsertNextCell(3, intp);
		}
	      }
	    }
	  }

	  int flen = strlen(argv[2]) + 80;
	  char* fname = new char[flen];
	  memset(fname, 0, flen);
	  strcpy(fname, argv[2]);
	  if (fname[strlen(fname)-1] != '/')
	    strcat(fname, "/");
	  const char* demname = dem->Name();
	  char c;
	  while (c = *demname++)
	    if (!isspace(c))
	      fname[strlen(fname)] = c;
	  fname[strlen(fname)] = '.';
	  fname[strlen(fname)] = '0' + k;
	  fname[strlen(fname)] = '.';
	  fname[strlen(fname)] = '0' + l;

	  strcat(fname, ".smf");
	  printf("filename is %s\n", fname);

	  // generate smf file
	  FILE* smf = fopen(fname, "w");
	  for (int ip = 0; ip < fltpoints->GetNumberOfPoints(); ip++) {
	    float p[3];
	    fltpoints->GetPoint(ip, p);
	    fprintf(smf, "v %f %f %f\n", p[0], p[1], p[2]);
	  }
	  int npts, *pts;
	  for (cells->InitTraversal(); cells->GetNextCell(npts, pts); )
	    fprintf(smf, "f %d %d %d\n", pts[0]+1, pts[1]+1, pts[2]+1);
	  fclose(smf);

	  delete [] ptidx;
	  fltpoints->Delete();
	  cells->Delete();
	}
      }
    }
    else {
	printf("sorry, can only handle 7.5 minute or 1 degree USGS DEMs in this version.\n");
    }
}
